/*******************************************************************************
 * Copyright (c) 2021-2023, Bayerische Motoren Werke Aktiengesellschaft (BMW AG)
 *
 * This program and the accompanying materials are made
 * available under the terms of the Eclipse Public License 2.0
 * which is available at https://www.eclipse.org/legal/epl-2.0/
 *
 * SPDX-License-Identifier: EPL-2.0
 *******************************************************************************/

//-----------------------------------------------------------------------------
/** @file  poly_line.h */
//-----------------------------------------------------------------------------

#ifndef MANTLEAPI_COMMON_POLY_LINE_H
#define MANTLEAPI_COMMON_POLY_LINE_H

#include <MantleAPI/Common/pose.h>
#include <MantleAPI/Common/time_utils.h>

#include <optional>
#include <vector>

namespace mantle_api
{
/// This struct represents the point of a polygonal chain (polyline) trajectory specification
struct PolyLinePoint
{
  Pose pose{};                ///< Pose of the PolyLinePoint
  std::optional<Time> time{}; ///< Time specification of the PolyLinePoint

  /// @brief Equality comparison for PolyLinePoint.
  ///
  /// @param[in]  other The left-hand side value for the comparison
  /// @returns  true if the values of `this` exactly equal to the values of other.
  bool operator==(const PolyLinePoint& other) const
  {
    return other.time == time && other.pose == pose;
  }

  /// @brief Prints a human-readable representation of 'polyLinePoint' to 'os'.
  /// @param os             The output stream
  /// @param polyLinePoint  The point of a polygonal chain (polyline) trajectory specification
  /// @returns 'polyLinePoint' in a human-readable format
  friend std::ostream& operator<<(std::ostream& os, const PolyLinePoint& polyLinePoint);
};

/// @brief Prints a human-readable representation of 'polyLinePoint' to 'os'.
/// @param os             The output stream
/// @param polyLinePoint  The point of a polygonal chain (polyline) trajectory specification
/// @returns 'polyLinePoint' in a human-readable format
inline std::ostream& operator<<(std::ostream& os, const PolyLinePoint& polyLinePoint)
{
  os << polyLinePoint.pose;

  if (polyLinePoint.time.has_value())
  {
    os << ", time in ms " << polyLinePoint.time.value();
  }

  return os;
}

/// The list of polyline points
using PolyLine = std::vector<PolyLinePoint>;

}  // namespace mantle_api

#endif  // MANTLEAPI_COMMON_POLY_LINE_H